/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2022  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QFont>
#include <QDir>
#include <QSettings>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QPushButton>
#include <QSpacerItem>
#include <QToolButton>
#include <QTextEdit>
#include <QVBoxLayout>
#include <QPixmap>
#include <QIcon>
#include <QCloseEvent>
#include <algorithm>
#include <assert.h>

#include "hexeditwidget.h"
#include "inputwidget.h"
#include "settingsdialog.h"

#include "dostmanndialog.h"


TDostmannDialog::TDostmannDialog(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QDialog(const_cast<QWidget*>(parent),flags)
{
	QFrame								*separator_hline;
	QFrame								*separator_vline;
	QGridLayout							*code_layout;
	QGridLayout							*help_layout;
	QGridLayout							*dialog_layout;
	QHBoxLayout							*input_hlayout;
	QHBoxLayout							*button_hlayout;
	QLabel								*existing_code1_label;
	QLabel								*existing_code2_label;
	QLabel								*input_label;
	QLabel								*new_code1_label;
	QLabel								*new_code2_label;
	QLabel								*data_range_label;
	QLabel								*icon_label;
	QLabel								*help0_label;
	QLabel								*help1_label;
	QLabel								*help2_label;
	QLabel								*help3_label;
	QPushButton							*close_button;
	QSpacerItem							*input_hspacer;
	QSpacerItem							*button_hspacer;
	QToolButton							*settings_button;
	QFont								large_font;
	QSizePolicy							sizepolicy_preferred_minexp(QSizePolicy::Preferred, QSizePolicy::MinimumExpanding);
	QSizePolicy							sizepolicy_preferred_fixed(QSizePolicy::Preferred, QSizePolicy::Fixed);
	QSizePolicy							sizepolicy_minexp_exp(QSizePolicy::MinimumExpanding, QSizePolicy::Expanding);
	QDir								data_location;
	QString								data_path;
	int									cntr;
	int									row_count;
	int									nominal_precision;
	int									measured_precision;
	double								nominal,actual,reading;
	static const int					MAX_ENTRY_WIDTH(100);

	this->resize(801,351);

	sizepolicy_minexp_exp.setHorizontalStretch(0);
	sizepolicy_minexp_exp.setVerticalStretch(0);

	sizepolicy_preferred_fixed.setHorizontalStretch(0);
	sizepolicy_preferred_fixed.setVerticalStretch(0);
	
	sizepolicy_preferred_minexp.setHorizontalStretch(0);
	sizepolicy_preferred_minexp.setVerticalStretch(0);

	data_location = QDir::home();
	
	if(!data_location.exists(".dostmann"))
	{
		data_location.mkdir(".dostmann");
	}
	
	data_location.cd(".dostmann");
	
	data_path = data_location.absolutePath();
	
	d_settings = new QSettings(data_path + "/Settings.ini",QSettings::IniFormat);

	large_font.setFamily(QStringLiteral("Verdana"));
	large_font.setPointSize(16);
	large_font.setWeight(57);
	
	dialog_layout = new QGridLayout(this);
	dialog_layout->setSizeConstraint(QLayout::SetFixedSize);
	
	input_hlayout = new QHBoxLayout();
	
	input_label = new QLabel(this);
	input_label->setFont(large_font);
	input_label->setAlignment(Qt::AlignLeft|Qt::AlignVCenter);
	input_hlayout->addWidget(input_label);

	input_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	input_hlayout->addItem(input_hspacer);
	
	settings_button = new QToolButton(this);
	settings_button->setFixedSize(QSize(22,22));
	input_hlayout->addWidget(settings_button);
	
	dialog_layout->addLayout(input_hlayout,0,0,1,1);
	
	d_input_widget = new TInputWidget(this);
	dialog_layout->addWidget(d_input_widget,1,0,1,1);

	help_layout = new QGridLayout();
	help_layout->setVerticalSpacing(1);
	
	help0_label = new QLabel(this);
	help_layout->addWidget(help0_label,0,0,1,1);
	
	help1_label = new QLabel(this);
	help_layout->addWidget(help1_label,1,0,1,1);
	
	help2_label = new QLabel(this);
	help_layout->addWidget(help2_label,2,0,1,1);
	
	help3_label = new QLabel(this);
	help_layout->addWidget(help3_label,3,0,1,1);

	dialog_layout->addLayout(help_layout,2,0,1,1);

	separator_vline = new QFrame(this);
	separator_vline->setFrameShape(QFrame::VLine);
	separator_vline->setFrameShadow(QFrame::Sunken);
	dialog_layout->addWidget(separator_vline,0,1,3,1);

	code_layout = new QGridLayout();
	
	existing_code1_label = new QLabel(this);
	existing_code1_label->setFont(large_font);
	existing_code1_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	code_layout->addWidget(existing_code1_label,0,0,1,1);
	
	d_existing_code1_edit = new THexEditWidget(this);
	d_existing_code1_edit->setAlignment(Qt::AlignCenter);
	d_existing_code1_edit->Set_Font(large_font);
	d_existing_code1_edit->setMaximumWidth(MAX_ENTRY_WIDTH);
	code_layout->addWidget(d_existing_code1_edit,0,1,1,1);

	existing_code2_label = new QLabel(this);
	existing_code2_label->setFont(large_font);
	existing_code2_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	code_layout->addWidget(existing_code2_label,1,0,1,1);
		
	d_existing_code2_edit = new THexEditWidget(this);
	d_existing_code2_edit->setAlignment(Qt::AlignCenter);
	d_existing_code2_edit->Set_Font(large_font);
	d_existing_code2_edit->setMaximumWidth(MAX_ENTRY_WIDTH);
	code_layout->addWidget(d_existing_code2_edit,1,1,1,1);

	new_code1_label = new QLabel(this);
	new_code1_label->setFont(large_font);
	new_code1_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	code_layout->addWidget(new_code1_label,2,0,1,1);

	d_new_code1_label = new QLabel(this);
	d_new_code1_label->setMinimumSize(100,0);
	d_new_code1_label->setFont(large_font);
	d_new_code1_label->setAlignment(Qt::AlignCenter);
	d_new_code1_label->setMaximumWidth(MAX_ENTRY_WIDTH);
	d_new_code1_label->setFrameShape(QFrame::StyledPanel);
	code_layout->addWidget(d_new_code1_label,2,1,1,1);

	new_code2_label = new QLabel(this);
	new_code2_label->setFont(large_font);
	new_code2_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	code_layout->addWidget(new_code2_label,3,0,1,1);

	d_new_code2_label = new QLabel(this);
	d_new_code2_label->setMinimumSize(100,0);
	d_new_code2_label->setFont(large_font);
	d_new_code2_label->setAlignment(Qt::AlignCenter);
	d_new_code2_label->setMaximumWidth(MAX_ENTRY_WIDTH);
	d_new_code2_label->setFrameShape(QFrame::StyledPanel);
	code_layout->addWidget(d_new_code2_label,3,1,1,1);
	
	data_range_label = new QLabel(this);
	data_range_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	code_layout->addWidget(data_range_label,4,0,1,1);
	
	d_data_range_label = new QLabel(this);
	d_data_range_label->setAlignment(Qt::AlignCenter);
	d_data_range_label->setMaximumWidth(MAX_ENTRY_WIDTH);
	code_layout->addWidget(d_data_range_label,4,1,1,1);
	
	icon_label = new QLabel(this);
	sizepolicy_preferred_minexp.setHeightForWidth(icon_label->sizePolicy().hasHeightForWidth());
	icon_label->setSizePolicy(sizepolicy_preferred_minexp);
	code_layout->addWidget(icon_label,5,0,1,1);
	
	d_toggle_help = new QToolButton(this);
	d_toggle_help->setCheckable(true);
	sizepolicy_preferred_fixed.setHeightForWidth(d_toggle_help->sizePolicy().hasHeightForWidth());
	d_toggle_help->setSizePolicy(sizepolicy_preferred_fixed);
	code_layout->addWidget(d_toggle_help,5,1,1,1);

	dialog_layout->addLayout(code_layout,0,2,3,1);
	
	d_help_edit = new QTextEdit(this);
	d_help_edit->setReadOnly(true);
	d_help_edit->setMinimumWidth(500);
	sizepolicy_minexp_exp.setHeightForWidth(d_help_edit->sizePolicy().hasHeightForWidth());
	d_help_edit->setSizePolicy(sizepolicy_minexp_exp);
	dialog_layout->addWidget(d_help_edit,0,3,5,1);

	separator_hline = new QFrame(this);
	separator_hline->setFrameShape(QFrame::HLine);
	separator_hline->setFrameShadow(QFrame::Sunken);
	dialog_layout->addWidget(separator_hline,3,0,1,3);

	button_hlayout = new QHBoxLayout();

	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);

	close_button = new QPushButton(this);
	close_button->setAutoDefault(false);
	button_hlayout->addWidget(close_button);
	dialog_layout->addLayout(button_hlayout,4,0,1,3);
	
	// set defaults
	d_settings_dialog = 0;
	d_help_edit->setVisible(false);
	row_count = d_settings->value(QStringLiteral("Row_Count"),3).toInt();
	
	if(row_count < 1)
	{
		row_count = 1;
	}
	
	d_input_widget->Clear();

	nominal_precision = d_settings->value(QStringLiteral("Nominal_Precision"),0).toInt();
	measured_precision = d_settings->value(QStringLiteral("Measured_Precision"),3).toInt();

	d_input_widget->Set_Precision(nominal_precision,measured_precision);
	
	for(cntr = 0;cntr < row_count;++cntr)
	{
		if(d_settings->contains(QString("Nominal_Value_%1").arg(cntr)))
		{
			nominal = d_settings->value(QString("Nominal_Value_%1").arg(cntr),20.0).toDouble();
			actual = d_settings->value(QString("Actual_Value_%1").arg(cntr),20.0).toDouble();
			reading = d_settings->value(QString("Reading_Value_%1").arg(cntr),20.0).toDouble();
			
			d_input_widget->Add_Entry(nominal,actual,reading);
		}
		else
		{
			nominal = static_cast<double>(15 + cntr * 5);
			actual = nominal;
			reading = nominal;
			
			d_input_widget->Add_Entry(nominal,actual,reading);
		}
	}
	
	d_existing_code1_edit->Set_Value(d_settings->value(QStringLiteral("Existing_Code1"),1.0).toDouble(),10000);
	d_existing_code2_edit->Set_Value(d_settings->value(QStringLiteral("Existing_Code2"),0.0).toDouble(),1000);
	
	this->Update_Codes();

	this->setWindowTitle(QStringLiteral("Dostmann Calibrator Version 2.1"));
	
	icon_label->setPixmap(QPixmap(":/images/meter.png"));
	settings_button->setIcon(QIcon(":/gui/settings_22x22.png"));

	input_label->setText(QStringLiteral("Input Temperaure Samples:"));
	existing_code1_label->setText(QStringLiteral("Existing Code 1:"));
	existing_code2_label->setText(QStringLiteral("Existing Code 2:"));
	new_code1_label->setText(QStringLiteral("New Code 1:"));
	new_code2_label->setText(QStringLiteral("New Code 2:"));
	data_range_label->setText(QStringLiteral("Data Range:"));
	d_toggle_help->setText(QStringLiteral("Show Help"));
	close_button->setText(QStringLiteral("Close"));
	
	help0_label->setText(QStringLiteral("Where:"));
	help1_label->setText(QStringLiteral("<html><head/><body><p><span style=\" font-weight:600;\">Nominal</span> is the nominal test temperature.</p></body></html>"));
	help2_label->setText(QStringLiteral("<html><head/><body><p><span style=\" font-weight:600;\">Actual</span> is the known or reference temperature of the sample.</p></body></html>"));
	help3_label->setText(QStringLiteral("<html><head/><body><p><span style=\" font-weight:600;\">Reading</span> is the temperature reported from the thermometer under test.</p></body></html>"));
	
	d_help_edit->setHtml(QStringLiteral("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
		"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
		"p, li { white-space: pre-wrap; }\n"
		"</style></head><body style=\" font-family:'.SF NS Text'; font-size:13pt; font-weight:400; font-style:normal;\">\n"
		"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:14pt; font-weight:600; text-decoration: underline;\">Recommended Calibration Procedure</span></p>\n"
		"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-size:14pt; font-weight:600; text-decoration: underline;\"><br /></p>\n"
		"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">The following is the recommended sequence for calibrating the Dostmann PT700 series thermometer"
		" specific for SCI:</p>\n"
		"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p>\n"
		"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600;\">1)</span> Update the entries for <span style=\" font-style:italic;\">Existing Code 1</span> and <span style=\" font-style:italic;\">Existing Code 2</span> with the current calibration codes from the Dostmann thermometer.</p>\n"
		"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p>\n"
		"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-style:italic; color:#0000ff;\">Note:  Non-calibrated values are 2710 for Code 1 and 0000 for Code 2.  Code 1 is the correction slope while Code 2 is th"
		"e correction offset.</span></p>\n"
		"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-style:italic; color:#0000ff;\"><br /></p>\n"
		"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600; color:#000000;\">2)</span><span style=\" color:#000000;\"> Configure the </span><span style=\" font-style:italic; color:#000000;\">Input Temperature Samples</span><span style=\" color:#000000;\"> for three entries with nominal temperatures of 15, 20, and 25 C.</span></p>\n"
		"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; color:#000000;\"><br /></p>\n"
		"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600; color:#000000;\">3)</span><span styl"
		"e=\" color:#000000;\"> Update the </span><span style=\" font-style:italic; color:#000000;\">Actual</span><span style=\" color:#000000;\"> and </span><span style=\" font-style:italic; color:#000000;\">Reading</span><span style=\" color:#000000;\"> entries for all displayed measurement samples.  New calibration codes will be generated automatically in the fields </span><span style=\" font-style:italic; color:#000000;\">New Code 1</span><span style=\" color:#000000;\"> and </span><span style=\" font-style:italic; color:#000000;\">New Code 2.</span></p>\n"
		"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; color:#000000;\"><br /></p>\n"
		"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600; color:#000000;\">4)</span><span style=\" color:#000000;\"> Enter the new correction codes in the Dostmann thermometer and update the </s"
		"pan><span style=\" font-style:italic; color:#000000;\">Existing Code 1 </span><span style=\" color:#000000;\">and </span><span style=\" font-style:italic; color:#000000;\">Existing Code 2</span><span style=\" color:#000000;\"> fields with the new correction values.</span></p>\n"
		"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; color:#000000;\"><br /></p>\n"
		"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600; color:#000000;\">5)</span><span style=\" color:#000000;\"> Configure the </span><span style=\" font-style:italic; color:#000000;\">Input Temperature Samples</span><span style=\" color:#000000;\"> for a single measurement entry by deleting two of the three entries.</span></p>\n"
		"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-ind"
		"ent:0px; color:#000000;\"><br /></p>\n"
		"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600; color:#000000;\">6)</span><span style=\" color:#000000;\"> Update the </span><span style=\" font-style:italic; color:#000000;\">Actual</span><span style=\" color:#000000;\"> and </span><span style=\" font-style:italic; color:#000000;\">Reading</span><span style=\" color:#000000;\"> entries for the single temperature sample.  The entry for </span><span style=\" font-style:italic; color:#000000;\">New Code 2</span><span style=\" color:#000000;\"> will be updated from this single temperature sample.</span></p>\n"
		"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-style:italic; color:#0000ff;\"><br /></p>\n"
		"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\""
		"><span style=\" font-weight:600; color:#000000;\">7)</span><span style=\" color:#000000;\"> Enter the new code 2 correction in the Dostmann thermometer and update the </span><span style=\" font-style:italic; color:#000000;\">Existing Code 2</span><span style=\" color:#000000;\"> field with the </span><span style=\" font-style:italic; color:#000000;\">New Code 2</span><span style=\" color:#000000;\"> value.</span></p></body></html>"));


	connect(d_input_widget,&TInputWidget::Data_Changed,this,&TDostmannDialog::Data_Changed);
	connect(d_existing_code1_edit,&THexEditWidget::Changed,this,&TDostmannDialog::Data_Changed);
	connect(d_existing_code2_edit,&THexEditWidget::Changed,this,&TDostmannDialog::Data_Changed);
	connect(settings_button,&QToolButton::clicked,this,&TDostmannDialog::Settings);
	connect(d_toggle_help,&QToolButton::toggled,this,&TDostmannDialog::Toggle_Help);
	connect(close_button,&QPushButton::clicked,this,&TDostmannDialog::close);
	
	this->Update_Codes();
}

TDostmannDialog::~TDostmannDialog(void)
{
}

void TDostmannDialog::closeEvent(
	QCloseEvent							*event)
{
	int									cntr;
	std::vector<TInputWidget::TEntryLineData> entry_data;
	std::vector<TInputWidget::TEntryLineData>::const_iterator entry_iter;
	
	entry_data = d_input_widget->Get_Data();
	
	d_settings->setValue(QStringLiteral("Row_Count"),static_cast<int>(entry_data.size()));
	
	for(cntr = 0,entry_iter = entry_data.begin();entry_iter != entry_data.end();++entry_iter,++cntr)
	{
		d_settings->setValue(QString("Nominal_Value_%1").arg(cntr),(*entry_iter).nominal);
		d_settings->setValue(QString("Actual_Value_%1").arg(cntr),(*entry_iter).reading);
		d_settings->setValue(QString("Reading_Value_%1").arg(cntr),(*entry_iter).reading);
	}
	
	d_settings->setValue(QStringLiteral("Existing_Code1"),d_existing_code1_edit->Value(10000));
	d_settings->setValue(QStringLiteral("Existing_Code2"),d_existing_code2_edit->Value(1000));

	d_settings->setValue(QStringLiteral("Nominal_Precision"),d_input_widget->Nominal_Precision());
	d_settings->setValue(QStringLiteral("Measured_Precision"),d_input_widget->Measured_Precision());

	event->accept();
}

void TDostmannDialog::Data_Changed(void)
{
	this->Update_Codes();
}

void TDostmannDialog::Settings(void)
{
	int									nominal_precision;
	int									measured_precision;
	
	if(!d_settings_dialog)
	{
		d_settings_dialog = new TSettingsDialog(this);
	}
	
	nominal_precision = d_input_widget->Nominal_Precision();
	measured_precision = d_input_widget->Measured_Precision();
	
	d_settings_dialog->Reset(nominal_precision,measured_precision);
	
	if(QDialog::Accepted == d_settings_dialog->exec())
	{		
		nominal_precision = d_settings_dialog->Nominal_Precision();
		measured_precision = d_settings_dialog->Measured_Precision();
		
		d_input_widget->Set_Precision(nominal_precision,measured_precision);
	}
}

void TDostmannDialog::Toggle_Help(
	bool								state)
{
	if(state)
	{
		d_toggle_help->setText(QStringLiteral("Hide Help"));
		
		d_help_edit->setVisible(true);
	}
	else
	{
		d_toggle_help->setText(QStringLiteral("Show Help"));
		d_help_edit->setVisible(false);
	}
}

void TDostmannDialog::Update_Codes(void)
{
	std::vector<TVector2>				pnts;
	std::vector<TVector2>::iterator		pnt_iter;
	std::vector<TInputWidget::TEntryLineData> entry_data;
	std::vector<TInputWidget::TEntryLineData>::const_iterator entry_iter;
	TVector2							pnt;
	double								slope;
	double								intercept;
	double								ref_min(0),ref_max(0);
	double								act_min(0),act_max(0);
	double								dval;
	bool								init;
	const double						zero_epsilon(0.000001);
	unsigned int						code1,code2;
	
	entry_data = d_input_widget->Get_Data();

	init = true;
	for(entry_iter = entry_data.begin();entry_iter != entry_data.end();++entry_iter)
	{
		pnt.x = (*entry_iter).reading;
		pnt.y = (*entry_iter).actual - (*entry_iter).reading;
		pnts.push_back(pnt);
		
		if(init)
		{
			ref_min = ref_max = (*entry_iter).actual;
			act_min = act_max = (*entry_iter).reading;
			init = false;
		}
		else
		{
			if((*entry_iter).actual > ref_max)
			{
				ref_max = (*entry_iter).actual;
			}
			else if((*entry_iter).actual < ref_min)
			{
				ref_min = (*entry_iter).actual;
			}
			
			if((*entry_iter).reading > act_max)
			{
				act_max = (*entry_iter).reading;
			}
			else if((*entry_iter).reading < act_min)
			{
				act_min = (*entry_iter).reading;
			}
		}
	}
	
	d_new_code1_label->setText(QStringLiteral("----"));
	d_new_code2_label->setText(QStringLiteral("----"));
	d_data_range_label->setText(QStringLiteral("---- ˚C"));
	
	// special case for single input
	// in this case ref_max/min and act_max/min are identical
	// the slope is nominal or the existing slope if enabled
	if(entry_data.size() == 1)
	{
		slope = 0.0;
		intercept = ref_max - act_max;
		
		slope += (d_existing_code1_edit->Value(10000) - 1.0);
		intercept += d_existing_code2_edit->Value(1000);

		code1 = 0xffff & static_cast<int>((slope + 1.0) * 10000.0);
		code2 = 0xffff & static_cast<int>(intercept * 1000.0);
			
		d_new_code1_label->setText(QString("%1").arg(QString("%1").arg(code1,4,16,QChar('0')).toUpper()));
		d_new_code2_label->setText(QString("%1").arg(QString("%1").arg(code2,4,16,QChar('0')).toUpper()));
	
		d_data_range_label->setText(QString("%1 ˚C").arg(0.0,0,'f',3));
	}
	else
	{
	
		if((ref_max - ref_min) > zero_epsilon &&
		   (act_max - act_min) > zero_epsilon &&
		   this->Fit_Line(pnts,&slope,&intercept))
		{			
			slope += (d_existing_code1_edit->Value(10000) - 1.0);
			intercept += d_existing_code2_edit->Value(1000);
			
			code1 = 0xffff & static_cast<int>((slope + 1.0) * 10000.0);
			code2 = 0xffff & static_cast<int>(intercept * 1000.0);
			
			d_new_code1_label->setText(QString("%1").arg(QString("%1").arg(code1,4,16,QChar('0')).toUpper()));
			d_new_code2_label->setText(QString("%1").arg(QString("%1").arg(code2,4,16,QChar('0')).toUpper()));
			
			init = true;

			for(pnt_iter = pnts.begin();pnt_iter != pnts.end();++pnt_iter)
			{
				dval = (*pnt_iter).y;
					
				// remove intercept
				dval -= intercept;
				
				if(init)
				{
					act_min = act_max = dval;
					init = false;
				}
				else
				{
					if(dval > act_max)
					{
						act_max = dval;
					}
					else if(dval < act_min)
					{
						act_min = dval;
					}
				}
			}
			
			d_data_range_label->setText(QString("%1 ˚C").arg(act_max - act_min,0,'f',3));
		}
	}
}

bool TDostmannDialog::Fit_Line(
	const std::vector<TVector2>			&pnts,
	double 								*slope,
	double 								*yintercept)
{
	double								XY,X2,Y2,Xs,Ys;
	int									point_count;
	std::vector<TVector2>::const_iterator 	pnt_iter;
	
	point_count = static_cast<int>(pnts.size());
	
	if(point_count < 2)
	{
		return false;
	}
	
	assert(slope);
	
	XY = X2 = Y2 = Xs = Ys =0;
	
	for(pnt_iter = pnts.begin();pnt_iter != pnts.end();pnt_iter++)
	{
		XY += ((*pnt_iter).x * (*pnt_iter).y);
		X2 += ((*pnt_iter).x * (*pnt_iter).x);
		Y2 += ((*pnt_iter).y * (*pnt_iter).y);
		Xs += (*pnt_iter).x;
		Ys += (*pnt_iter).y;
	}
	
	*slope = static_cast<double>(point_count) * XY - Xs * Ys;
	*slope /= (static_cast<double>(point_count) * (X2) - Xs * Xs);
	
	if(yintercept)
	{
		*yintercept = Ys - (*slope) * Xs;
		*yintercept /= static_cast<double>(point_count);
	}
	
	return true;
}


