/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2022  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QCheckBox>
#include <QToolButton>
#include <QSpacerItem>
#include <QSize>
#include <assert.h>

#include "inputwidget.h"

TInputWidget::TInputWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QLabel								*actual_label;
	QLabel								*correction_label;
	QLabel								*nominal_label;
	QLabel								*reading_label;
	QSpacerItem							*button_hspacer;
	
	d_nominal_precision = 0;
	d_measured_precision = 2;

	d_grid_layout = new QGridLayout(this);
	d_grid_layout->setSpacing(3);

	nominal_label = new QLabel(this);
	nominal_label->setAlignment(Qt::AlignCenter);
	d_grid_layout->addWidget(nominal_label,0,1,1,1);

	actual_label = new QLabel(this);
	actual_label->setAlignment(Qt::AlignCenter);
	d_grid_layout->addWidget(actual_label,0,2,1,1);

	reading_label = new QLabel(this);
	reading_label->setAlignment(Qt::AlignCenter);
	d_grid_layout->addWidget(reading_label,0,3,1,1);

	correction_label = new QLabel(this);
	correction_label->setAlignment(Qt::AlignCenter);
	d_grid_layout->addWidget(correction_label,0,4,1,1);
	
	d_widget_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	d_grid_layout->addItem(d_widget_vspacer,1,0,1,5);

	d_button_hlayout = new QHBoxLayout();

	d_add_button = new QToolButton(this);
	d_add_button->setFixedSize(QSize(22,22));
	d_button_hlayout->addWidget(d_add_button);
	
	d_remove_button = new QToolButton(this);
	d_remove_button->setFixedSize(QSize(22,22));
	d_button_hlayout->addWidget(d_remove_button);
	
	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	d_button_hlayout->addItem(button_hspacer);
	
	d_grid_layout->addLayout(d_button_hlayout,2,0,1,5);
		
	d_add_button->setIcon(QIcon(":/gui/add_22x22.png"));
	d_remove_button->setIcon(QIcon(":/gui/remove_22x22.png"));

	nominal_label->setText(QStringLiteral("Nominal"));
	actual_label->setText(QStringLiteral("Actual"));
	reading_label->setText(QStringLiteral("Reading"));
	correction_label->setText(QStringLiteral("Correction\n(Error)"));
	
	// defaults
	d_remove_button->setEnabled(false);
	
	connect(d_add_button,&QToolButton::clicked,this,&TInputWidget::Add_Row);
	connect(d_remove_button,&QToolButton::clicked,this,&TInputWidget::Remove_Rows);
}

TInputWidget::~TInputWidget(void)
{
}

std::vector<TInputWidget::TEntryLineData> TInputWidget::Get_Data(void) const
{
	std::vector<TInputWidget::TEntryLine>::const_iterator iter;
	std::vector<TInputWidget::TEntryLineData> entry_data;
	TInputWidget::TEntryLineData		entry;
	
	for(iter = d_entry_lines.begin();iter != d_entry_lines.end();++iter)
	{
		entry.nominal = (*iter).nominal_edit->text().toDouble();
		entry.actual = (*iter).actual_edit->text().toDouble();
		entry.reading = (*iter).reading_edit->text().toDouble();
		entry.correction = (*iter).correction_edit->text().toDouble();
		
		entry_data.push_back(entry);
	}
	
	return entry_data;
}

TInputWidget::TEntryLineData TInputWidget::Get_Data(
	const int							index) const
{
	TInputWidget::TEntryLineData		entry;
	
	assert(static_cast<unsigned int>(index) < d_entry_lines.size());
	
	entry.nominal = d_entry_lines[index].nominal_edit->text().toDouble();
	entry.actual = d_entry_lines[index].actual_edit->text().toDouble();
	entry.reading = d_entry_lines[index].reading_edit->text().toDouble();
	entry.correction = d_entry_lines[index].correction_edit->text().toDouble();
	
	return entry;
}

void TInputWidget::Clear(void)
{
	this->Clear_Entries();
	d_entry_lines.clear();
}

void TInputWidget::Add_Entry(
	const double						&nominal,
	const double						&actual,
	const double						&reading)
{
	int									row;
	
	row = this->Create_Entry_Line();
	this->Set_Entry_Line_Data(row,nominal,actual,reading);
	
	emit Data_Changed();
}

void TInputWidget::Set_Precision(
	int									nominal_precision,
	int									measured_precision)
{
	std::vector<TInputWidget::TEntryLine>::iterator iter;
	double								dval;
	double								actual;
	double								reading;
	bool								state;
	
	if(static_cast<unsigned int>(nominal_precision) < 3)
	{
		d_nominal_precision = nominal_precision;
	}
	
	if(static_cast<unsigned int>(measured_precision) < 5)
	{
		d_measured_precision = measured_precision;
	}
	
	for(iter = d_entry_lines.begin();iter != d_entry_lines.end();++iter)
	{
		dval = (*iter).nominal_edit->text().toDouble(&state);
		if(state) (*iter).nominal_edit->setText(QString("%1").arg(dval,0,'f',d_nominal_precision));

		actual = (*iter).actual_edit->text().toDouble(&state);
		if(state) (*iter).actual_edit->setText(QString("%1").arg(actual,0,'f',d_measured_precision));
		
		reading = (*iter).reading_edit->text().toDouble(&state);
		if(state) (*iter).reading_edit->setText(QString("%1").arg(reading,0,'f',d_measured_precision));
		
		(*iter).correction_edit->setText(QString("%1").arg(reading-actual,0,'f',d_measured_precision));
	}
}

void TInputWidget::Row_Selection_Changed(void)
{
	std::vector<TInputWidget::TEntryLine>::const_iterator iter;
	bool								state(false);
	
	for(iter = d_entry_lines.begin();iter != d_entry_lines.end();++iter)
	{
		if((*iter).line_check->isChecked())
		{
			state = true;
			break;
		}
	}
	
	d_remove_button->setEnabled(state);
}

void TInputWidget::Row_Data_Changed(void)
{
	double								actual;
	double								reading;
	std::vector<TInputWidget::TEntryLine>::iterator iter;
	bool								state;
	
	for(iter = d_entry_lines.begin();iter != d_entry_lines.end();++iter)
	{
		actual = (*iter).actual_edit->text().toDouble(&state);
		
		if(state)
		{
			reading = (*iter).reading_edit->text().toDouble(&state);
			
			if(state)
			{
				(*iter).correction_edit->setText(QString("%1").arg(reading-actual,0,'f',d_measured_precision));
			}
		}
	}
	
	emit Data_Changed();
}

void TInputWidget::Add_Row(void)
{
	int									row;
	
	row = this->Create_Entry_Line();
	
	this->Set_Entry_Line_Data(row,20.0,20.0,20.0);
	
	emit Data_Changed();
}

void TInputWidget::Remove_Rows(void)
{
	std::vector<TInputWidget::TEntryLine>::iterator iter;
	std::vector<TInputWidget::TEntryLineData> entry_data;
	std::vector<TInputWidget::TEntryLineData>::iterator entry_iter;
	TInputWidget::TEntryLineData		entry;
	int									row;

	// backup and clear
	for(iter = d_entry_lines.begin();iter != d_entry_lines.end();++iter)
	{
		if(!(*iter).line_check->isChecked())
		{
			entry.nominal = (*iter).nominal_edit->text().toDouble();
			entry.actual = (*iter).actual_edit->text().toDouble();
			entry.reading = (*iter).reading_edit->text().toDouble();
			entry.correction = (*iter).correction_edit->text().toDouble();
			
			entry_data.push_back(entry);
		}
	}
	
	this->Clear_Entries();
	d_entry_lines.clear();
	
	// make sure at least one entry exists
	if(entry_data.size() < 1)
	{
		entry.nominal = 20.0;
		entry.actual = 20.0;
		entry.reading = 20.0;
		entry.correction = 0.0;
		
		entry_data.push_back(entry);
	}
	
	// rebuild
	for(entry_iter = entry_data.begin();entry_iter != entry_data.end();++entry_iter)
	{
		row = this->Create_Entry_Line();
		this->Set_Entry_Line_Data(row,(*entry_iter).nominal,(*entry_iter).actual,(*entry_iter).reading);
	}
	
	d_remove_button->setEnabled(false);
	
	emit Data_Changed();
}

int TInputWidget::Create_Entry_Line(void)
{
	TInputWidget::TEntryLine			entry_line;
	int									row;
	static const int					MAX_ENTRY_WIDTH(90);
	
	assert(d_grid_layout);
	assert(d_button_hlayout);
	assert(d_widget_vspacer);
	
	d_grid_layout->removeItem(d_button_hlayout);
	d_grid_layout->removeItem(d_widget_vspacer);
	
	row = 1 + static_cast<int>(d_entry_lines.size());

	entry_line.line_check = new QCheckBox(this);
	entry_line.line_check->setMaximumWidth(MAX_ENTRY_WIDTH);
	d_grid_layout->addWidget(entry_line.line_check,row,0,1,1);

	entry_line.nominal_edit = new QLineEdit(this);
	entry_line.nominal_edit->setAlignment(Qt::AlignCenter);
	entry_line.nominal_edit->setMaximumWidth(MAX_ENTRY_WIDTH);
	d_grid_layout->addWidget(entry_line.nominal_edit,row,1,1,1);
	
	entry_line.actual_edit = new QLineEdit(this);
	entry_line.actual_edit->setAlignment(Qt::AlignCenter);
	entry_line.actual_edit->setMaximumWidth(MAX_ENTRY_WIDTH);
	d_grid_layout->addWidget(entry_line.actual_edit,row,2,1,1);
	
	entry_line.reading_edit = new QLineEdit(this);
	entry_line.reading_edit->setAlignment(Qt::AlignCenter);
	entry_line.reading_edit->setMaximumWidth(MAX_ENTRY_WIDTH);
	d_grid_layout->addWidget(entry_line.reading_edit,row,3,1,1);
	
	entry_line.correction_edit = new QLineEdit(this);
	entry_line.correction_edit->setAlignment(Qt::AlignCenter);
	entry_line.correction_edit->setReadOnly(true);
	entry_line.correction_edit->setMaximumWidth(MAX_ENTRY_WIDTH);
	d_grid_layout->addWidget(entry_line.correction_edit,row,4,1,1);
	
	entry_line.units_label = new QLabel(this);
	d_grid_layout->addWidget(entry_line.units_label,row,5,1,1);

	entry_line.line_check->setText(QStringLiteral(" "));
	entry_line.units_label->setText(QStringLiteral("˚C"));
	
	d_entry_lines.push_back(entry_line);
	
	d_grid_layout->addItem(d_widget_vspacer,row + 1,0,1,5);
	d_grid_layout->addLayout(d_button_hlayout,row + 2,0,1,5);
	
	connect(entry_line.line_check,&QCheckBox::toggled,this,&TInputWidget::Row_Selection_Changed);
	
	connect(entry_line.actual_edit,&QLineEdit::editingFinished,this,&TInputWidget::Row_Data_Changed);
	connect(entry_line.reading_edit,&QLineEdit::editingFinished,this,&TInputWidget::Row_Data_Changed);
	
	return (static_cast<int>(d_entry_lines.size()) - 1);
}

void TInputWidget::Set_Entry_Line_Data(
	const int							row,
	const double						&nominal,
	const double						&actual,
	const double						&reading)
{
	bool								prev_state;
	
	assert(row < static_cast<int>(d_entry_lines.size()));
	
	prev_state = d_entry_lines[row].nominal_edit->blockSignals(true);
	d_entry_lines[row].nominal_edit->setText(QString("%1").arg(nominal,0,'f',d_nominal_precision));
	d_entry_lines[row].nominal_edit->blockSignals(prev_state);
	
	prev_state = d_entry_lines[row].actual_edit->blockSignals(true);
	d_entry_lines[row].actual_edit->setText(QString("%1").arg(actual,0,'f',d_measured_precision));
	d_entry_lines[row].actual_edit->blockSignals(prev_state);

	prev_state = d_entry_lines[row].reading_edit->blockSignals(true);
	d_entry_lines[row].reading_edit->setText(QString("%1").arg(reading,0,'f',d_measured_precision));
	d_entry_lines[row].reading_edit->blockSignals(prev_state);

	d_entry_lines[row].correction_edit->setText(QString("%1").arg(reading-actual,0,'f',d_measured_precision));
}

void TInputWidget::Clear_Entries(void)
{
	std::vector<TInputWidget::TEntryLine>::iterator iter;
	
	assert(d_grid_layout);
	assert(d_button_hlayout);
	assert(d_widget_vspacer);
	
	d_grid_layout->removeItem(d_button_hlayout);
	d_grid_layout->removeItem(d_widget_vspacer);
	
	for(iter = d_entry_lines.begin();iter != d_entry_lines.end();++iter)
	{
		d_grid_layout->removeWidget((*iter).line_check);
		d_grid_layout->removeWidget((*iter).nominal_edit);
		d_grid_layout->removeWidget((*iter).actual_edit);
		d_grid_layout->removeWidget((*iter).reading_edit);
		d_grid_layout->removeWidget((*iter).correction_edit);
		d_grid_layout->removeWidget((*iter).units_label);
		
		disconnect((*iter).line_check,0,0,0);
		disconnect((*iter).actual_edit,0,0,0);
		disconnect((*iter).reading_edit,0,0,0);
		
		(*iter).line_check->deleteLater();
		(*iter).nominal_edit->deleteLater();
		(*iter).actual_edit->deleteLater();
		(*iter).reading_edit->deleteLater();
		(*iter).correction_edit->deleteLater();
		(*iter).units_label->deleteLater();
	}
	
	d_grid_layout->addItem(d_widget_vspacer,1,0,1,5);
	d_grid_layout->addLayout(d_button_hlayout,2,0,1,5);
}

